/* This file is part of syslogrelay.
 * Copyright (C) 2022 Sergey Poznyakoff
 *
 * Syslogrelay is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * Syslogrelay is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with syslogrelay.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <sysexits.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>

char *progname;   /* This program name */
char *hostname;   /* This host name */
size_t hostname_length;
int debug;        /* Debug level */
int max_prio = LOG_DEBUG;  /* Log messages with priorities <= max_prio */
char *logdev = "/dev/log";

int skip_hostname = 0;

enum {
	FILE_PRIO_NONE,
	FILE_PRIO_KEEP,
	FILE_PRIO_DECODE
};

int file_prio = FILE_PRIO_NONE;

#define MAXLINE 1024

/* Output channel types. */
enum channel_type {
	CHAN_FILE,  /* Log to file. */
	CHAN_UDP    /* Relay to a remote collector. */
};

struct output_channel {
	enum channel_type type;
	char *name;
	int fd;
};

struct output_channel *channel;

static void
error(int status, int errnum, char const *fmt, ...)
{
	va_list ap;

	fprintf(stderr, "%s: ", progname);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	if (errnum)
		fprintf(stderr, ": %s", strerror(errnum));
	fputc('\n', stderr);
	if (status)
		exit(status);
}

#define PRI_FAC(p) ((p) >> 3)
#define NUM2FAC(n) ((n) << 3)
#define PRI_SEV(n) ((n) & 0x7)
#define NSTR(t) (sizeof(t)/sizeof(t[0]))

static char const *strfac[] = {
    [PRI_FAC(LOG_USER)] =      "USER",
    [PRI_FAC(LOG_MAIL)] =      "MAIL",
    [PRI_FAC(LOG_DAEMON)] =    "DAEMON",
    [PRI_FAC(LOG_AUTH)] =      "AUTH",
    [PRI_FAC(LOG_AUTHPRIV)] =  "AUTHPRIV",
    [PRI_FAC(LOG_SYSLOG)] =    "SYSLOG",
    [PRI_FAC(LOG_LPR)] =       "LPR",
    [PRI_FAC(LOG_NEWS)] =      "NEWS",
    [PRI_FAC(LOG_UUCP)] =      "UUCP",
    [PRI_FAC(LOG_FTP)] =       "FTP",
    [PRI_FAC(LOG_CRON)] =      "CRON",
    [PRI_FAC(LOG_LOCAL0)] =    "LOCAL0",
    [PRI_FAC(LOG_LOCAL1)] =    "LOCAL1",
    [PRI_FAC(LOG_LOCAL2)] =    "LOCAL2",
    [PRI_FAC(LOG_LOCAL3)] =    "LOCAL3",
    [PRI_FAC(LOG_LOCAL4)] =    "LOCAL4",
    [PRI_FAC(LOG_LOCAL5)] =    "LOCAL5",
    [PRI_FAC(LOG_LOCAL6)] =    "LOCAL6",
    [PRI_FAC(LOG_LOCAL7)] =    "LOCAL7",
};

static char const *strpri[] = {
    [LOG_EMERG] =   "EMERG",
    [LOG_ALERT] =   "ALERT",
    [LOG_CRIT] =    "CRIT",
    [LOG_ERR] =     "ERR",
    [LOG_WARNING] = "WARNING",
    [LOG_NOTICE] =  "NOTICE",
    [LOG_INFO] =    "INFO",
    [LOG_DEBUG] =   "DEBUG",
};

static inline int
kw_to_num(char const *kw, char const **tab, int len)
{
    int i;
    for (i = 0; i < len; i++)
	if (tab[i] && strcasecmp(kw, tab[i]) == 0)
	    return i;
    return -1;
}

static inline char const *
num_to_kw(int n, char const **tab, int len)
{
    if (n < 0 || n > len)
	return NULL;
    return tab[n];
}

static void
enomem(void)
{
	error(EX_OSERR, 0, "out of memory");
}

static void *
xmalloc(size_t size)
{
	void *p = malloc(size);
	if (!p)
		enomem();
	return p;
}

static char *
xstrdup(char *str)
{
	char *p = strdup(str);
	if (!p)
		enomem();
	return p;
}

#ifndef HOST_NAME_MAX
# define HOST_NAME_MAX 64
#endif

static char *
xgethostname(void)
{
	  char *result = NULL;
	  size_t size = 0;
	  char *p;

	  for (;;) {
		  if (size == 0) {
			  size = HOST_NAME_MAX;
		  } else if ((size_t) -1 / 3 * 2 <= size) {
			  enomem();
		  } else {
			  size += (size + 1) / 2;
		  }
		  p = realloc(result, size);
		  if (!p)
			  enomem();
		  result = p;
		  result[size - 1] = 0;
		  if (gethostname(result, size - 1) == 0) {
			  if (!result[size - 1])
				  break;
		  } else if (errno != 0 && errno != ENAMETOOLONG
			     && errno != EINVAL && errno != ENOMEM)
			  error(EX_OSERR, errno, "can't determine my host name");
	  }

	  /* Try to return fully qualified host name */
	  if (!strchr(result, '.')) {
		  struct addrinfo hints, *rp;

		  memset(&hints, 0, sizeof(hints));
		  hints.ai_family = AF_UNSPEC;
		  hints.ai_flags = AI_CANONNAME;

		  if (getaddrinfo(result, NULL, &hints, &rp) == 0) {
			  free(result);
			  result = xstrdup(rp->ai_canonname);
			  freeaddrinfo(rp);
		  }
	  }
	  return result;
}

#define	HAS_PRI        0x1
#define	HAS_TIMESTAMP  0x2

#define TIMESTAMP_LENGTH 15

struct message {
	int flags;
	int pri;
	int ts_start;         /* Start of the timestamp */
	char buf[MAXLINE];
	size_t len;
};

static int
ismonth(char const *buf, size_t len)
{
	static char *mon[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			       "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	int i;

	if (len <= 3 || buf[3] != ' ')
		return 0;
	for (i = 0; i < sizeof(mon)/sizeof(mon[0]); i++)
		if (memcmp(mon[i], buf, 3) == 0)
			return 1;
	return 0;
}

void
parse_message(struct message *msg, char *buf, size_t len)
{
	char *p;
	unsigned long n;
	int i;

	memset(msg, 0, sizeof(*msg));
	memcpy(msg->buf, buf, len);
	msg->len = len;

	if (len < 5 || *buf != '<')
		return;

	errno = 0;
	n = strtoul(buf + 1, &p, 10);
	if (errno)
		return;

	if ((p - buf) > 5 || *p != '>')
		return;

	msg->flags = HAS_PRI;
	msg->pri = n;

	/* Skip past the timestamp part */
	msg->ts_start = i = p - buf + 1;
	if (!ismonth(buf + i, len - i))
		return;
	i += 4;

	/* Skip the day of month */
	if (len - i > 3
	    && (buf[i] == ' ' || isdigit(buf[i])) && isdigit(buf[i+1])
	    && buf[i+2] == ' ')
		i += 3;
	else
		return;

	/* Skip HH:MM:SS */
	if (len - i > 9
	    && isdigit(buf[i])
	    && isdigit(buf[i+1])
	    && buf[i+2] == ':'
	    && isdigit(buf[i+3])
	    && isdigit(buf[i+4])
	    && buf[i+5] == ':'
	    && isdigit(buf[i+6])
	    && isdigit(buf[i+7])
	    && buf[i+8] == ' ')
		i += 9;
	else
		return;

	msg->flags |= HAS_TIMESTAMP;
}

void
message_insert(struct message *msg, char const *text, size_t text_len, size_t off)
{
	msg->len += text_len;
	if (msg->len > MAXLINE)
		msg->len = MAXLINE;
	memmove(msg->buf + off + text_len, msg->buf + off, msg->len - off - text_len);
	memcpy(msg->buf + off, text, text_len);
}

void
message_normalize(struct message *msg)
{
	int add_hostname;

	if (!(msg->flags & HAS_PRI)) {
		static char default_pri[] = "<13> ";
		static int default_pri_len = sizeof(default_pri) - 1;
		message_insert(msg, default_pri, default_pri_len, 0);
		msg->pri = 13;
		msg->ts_start = default_pri_len;
	}
	if (!(msg->flags & HAS_TIMESTAMP)) {
		struct timeval tv;
		struct tm *tm;
		char ts[TIMESTAMP_LENGTH+2];

		gettimeofday(&tv, NULL);
		tm = localtime(&tv.tv_sec);
		strftime(ts, sizeof(ts), "%b %e %T ", tm);
		message_insert(msg, ts, TIMESTAMP_LENGTH+1, msg->ts_start);
		add_hostname = 1;
	} else {
		add_hostname = (msg->len > hostname_length &&
				strncasecmp(msg->buf + msg->ts_start + TIMESTAMP_LENGTH + 1,
					    hostname, hostname_length));
	}

	if (add_hostname && !skip_hostname)
		message_insert(msg, hostname, hostname_length + 1,
			       msg->ts_start + TIMESTAMP_LENGTH + 1);
}

static int
open_unix_socket(const char *path)
{
	int fd;
	struct sockaddr_un sun;

	if (strlen(path) >= sizeof(sun.sun_path))
		error(EX_UNAVAILABLE, 0, "UNIX socket name too long: %s", path);

	unlink(path);

	memset(&sun, 0, sizeof(sun));
	sun.sun_family = AF_UNIX;
	strncpy(sun.sun_path, path, sizeof(sun.sun_path) - 1);

	fd = socket(AF_UNIX, SOCK_DGRAM, 0);
	if (fd < 0)
		error(EX_UNAVAILABLE, errno, "can't create unix socket");

	if (bind(fd, (struct sockaddr *) &sun, sizeof(sun)))
		error(EX_UNAVAILABLE, errno, "can't bind to %s", path);

	if (chmod(path, 0666))
		error(EX_UNAVAILABLE, errno, "can't chmod %s", path);

	return fd;
}

static int
chan_file_open(struct output_channel *chan)
{
	int fd;

	if (strcmp(chan->name, "-") == 0) {
		fd = dup(1);
		if (fd == -1)
			error(EX_CANTCREAT, errno, "can't dup stdout");
	} else {
		fd = open(chan->name, O_WRONLY | O_APPEND | O_CREAT, 0644);
		if (fd == -1)
			error(EX_CANTCREAT, errno, "can't open %s for writing", chan->name);
	}
	chan->fd = fd;
	return fd;
}

#define IOVCNT 7

static int
chan_file_send(struct output_channel *chan, struct message *msg)
{
	struct iovec iov[IOVCNT];
	struct iovec *v = iov;

	char *buf = msg->buf;
	size_t len = msg->len;

	switch (file_prio) {
	case FILE_PRIO_NONE:
		buf += msg->ts_start;
		len -= msg->ts_start;
		break;

	case FILE_PRIO_KEEP:
		break;

	case FILE_PRIO_DECODE: {
		char const *f;
		char const *p;

		f = num_to_kw(PRI_FAC(msg->pri), strfac, NSTR(strfac));
		p = num_to_kw(PRI_SEV(msg->pri), strpri, NSTR(strpri));
		if (f && p) {
			v->iov_base = "<";
			v->iov_len = 1;
			v++;

			v->iov_base = (char*) f;
			v->iov_len = strlen(f);
			v++;

			v->iov_base = ".";
			v->iov_len = 1;
			v++;

			v->iov_base = (char*) p;
			v->iov_len = strlen(p);
			v++;

			v->iov_base = ">";
			v->iov_len = 1;
			v++;

			buf += msg->ts_start;
			len -= msg->ts_start;
		}
	}
	}

	v->iov_base = (char *) buf;
	v->iov_len = len;
	v++;
	v->iov_base = (char *) "\n";
	v->iov_len = 1;
	v++;

	if (writev(chan->fd, iov, v - iov) < 0) {
		error(0, errno, "error writing to %s", chan->name);
		return -1;
	}
	return 0;
}

struct udp_output_channel {
	struct output_channel base;
	struct sockaddr *sa;
	socklen_t salen;
	int connected;
};

static struct output_channel *
chan_udp_create(void)
{
	struct udp_output_channel *chan = calloc(1, sizeof(*chan));
	if (!chan)
		enomem();
	return (struct output_channel *) chan;
}

static int
chan_udp_open(struct output_channel *chan)
{
	struct udp_output_channel *uchan = (struct udp_output_channel *)chan;
	int fd;
	struct addrinfo hints, *rp;
	char *p, *node, *service;
	int rc;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_protocol = IPPROTO_UDP;

	p = strrchr(chan->name, ':');
	if (p) {
		size_t len = p - chan->name;

		if (len == 0)
			error(EX_USAGE, 0, "invalid remote address: %s", chan->name);
		node = xmalloc(len + 1);
		memcpy(node, chan->name, len);
		node[len] = 0;

		service = strdup(p + 1);
	} else {
		node = strdup(chan->name);
		service = strdup("514");
		hints.ai_flags |= AI_NUMERICSERV;
	}
	if (node[0] == '[') {
		size_t len = strlen(node);
		if (node[len - 1] == ']') {
			hints.ai_family = AF_INET6;
			memmove(node, node + 1, len - 2);
			node[len - 2] = 0;
		}
	}
	rc = getaddrinfo(node, service, &hints, &rp);
	free(node);
	free(service);
	if (rc)
		error(EX_UNAVAILABLE, 0, "can't resolve %s: %s", chan->name, gai_strerror(rc));

	if (rp->ai_next)
		error(EX_USAGE, 0, "%s resolves to multiple addresses", chan->name);
	fd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
	if (fd < 0)
		error(EX_OSERR, errno, "can't create socket %s", chan->name);
	chan->fd = fd;
	uchan->sa = xmalloc(rp->ai_addrlen);
	memcpy(uchan->sa, rp->ai_addr, rp->ai_addrlen);
	uchan->salen = rp->ai_addrlen;
	freeaddrinfo(rp);

	if (connect(chan->fd, uchan->sa, uchan->salen) == 0) {
		uchan->connected = 1;
	} else {
		uchan->connected = 0;
		error(0, errno, "can't connect to %s", chan->name);
	}

	return fd;
}

static void
chan_udp_close(struct output_channel *chan)
{
	struct udp_output_channel *uchan = (struct udp_output_channel *)chan;
	close(chan->fd);
	chan->fd = -1;
	free(uchan->sa);
	uchan->sa = NULL;
	uchan->salen = 0;
	uchan->connected = 0;
}

static int
chan_udp_reconnect(struct udp_output_channel *uchan)
{
	if (connect(uchan->base.fd, uchan->sa, uchan->salen) == 0) {
		uchan->connected = 1;
	} else {
		uchan->connected = 0;
		error(0, errno, "can't connect to %s", uchan->base.name);
	}
	return uchan->connected;
}

static int
chan_udp_send(struct output_channel *chan, struct message *msg)
{
	struct udp_output_channel *uchan = (struct udp_output_channel *)chan;
	int rc;

	if (!uchan->connected) {
		if (!chan_udp_reconnect(uchan))
			return -1;
	}

	rc = send(chan->fd, msg->buf, msg->len, 0);
	if (rc == -1) {
		int ec = errno;

		error(0, errno, "send failed");

		if (ec == ECONNREFUSED || /* connection went down */
		    ec == ENOTCONN ||     /* nobody listening */
		    ec == EDESTADDRREQ || /* BSD equivalents of the above */
		    ec == ECONNRESET) {
			chan_udp_reconnect(uchan);
		}
		return -1;
	} else if (rc != (ssize_t) msg->len) {
		error(0, errno, "sent less bytes than expected (%lu vs. %lu)",
		      (unsigned long) rc, (unsigned long) msg->len);
		return -1;
	}
	return 0;
}

struct channel_method {
	struct output_channel *(*chan_create)(void);
	int (*chan_open)(struct output_channel *);
	int (*chan_send)(struct output_channel *, struct message *);
	void (*chan_close)(struct output_channel *);
};

static struct output_channel *
default_chan_create(void)
{
	return xmalloc(sizeof(struct output_channel));
}

static void
default_chan_close(struct output_channel *chan)
{
	close(chan->fd);
}

static struct channel_method channel_method[] = {
	{ default_chan_create, chan_file_open, chan_file_send, default_chan_close },
	{ chan_udp_create, chan_udp_open, chan_udp_send, chan_udp_close }
};

struct output_channel *
channel_open(int type, char *name)
{
	struct output_channel *chan;

	chan = channel_method[type].chan_create();
	chan->type = type;
	chan->name = xstrdup(name);
	chan->fd = -1;
	if (channel_method[type].chan_open(chan) == -1)
		error(EX_UNAVAILABLE, 0, "can't open output channel");

	return chan;
}

int
channel_send(struct output_channel *chan, struct message *msg)
{
	return channel_method[chan->type].chan_send(chan, msg);
}

void
channel_close(struct output_channel *chan)
{
	channel_method[chan->type].chan_close(chan);
	free(chan->name);
	free(chan);
}

static void
usage(void)
{
	printf("usage: %s [OPTION]\n", progname);
	printf("Relays syslog messages to a remote server or to a file.\n");
	printf("\nOPTIONS are:\n");
	printf("\nLog destination:\n");
	printf("    -O FILE         store messages on FILE\n");
	printf("    -R HOST[:PORT]  relay messages to HOST:PORT (UDP; PORT defaults to 514)\n");

	printf("\nHostname insertion:\n");
	printf("    -H HOSTNAME     use this hostname\n");
	printf("    -S              don't add hostname to the message\n");

	printf("\nPRI handling (with -O only):\n");
	printf("    -k              keep PRI part of each message on output\n");
        printf("    -D              decode PRI part of each message to a human-readable form\n");

	printf("\nOther options:\n");
	printf("    -d              increase debug verbosity\n");
        printf("    -h              print this help list\n");
	printf("    -l SEVERITY     skip messages with severities greater than the given one\n");
	printf("    -n              ignored\n");
	printf("    -p SOCKET       listen on this UNIX socket (default: /dev/log)\n");

	printf("\n");
}

enum {
	SIGNAL_RELOAD,
	SIGNAL_STOP
};

static int volatile got_signal; 

static void
sighan(int sig)
{
	switch (sig) {
	case SIGHUP:
		got_signal = SIGNAL_RELOAD;
		break;

	default:
		got_signal = SIGNAL_STOP;
	}
}

int
main(int argc, char **argv)
{
	int c;
	int fd;
	int chtype = CHAN_FILE;
	char *chname = "/var/log/messages";
	struct sigaction act;
	
	progname = strrchr(argv[0], '/');
	if (progname)
		++progname;
	else
		progname = argv[0];

	while ((c = getopt(argc, argv, "DH:O:SR:dhkl:np:")) != EOF) {
		switch (c) {
		case 'D':
			file_prio = FILE_PRIO_DECODE;
			break;

		case 'H':
			hostname = xstrdup(optarg);
			break;

		case 'O':
			chtype = CHAN_FILE;
			chname = optarg;
			break;

		case 'S':
			skip_hostname = 1;
			break;

		case 'R':
			chtype = CHAN_UDP;
			chname = optarg;
			break;

		case 'd':
			debug++;
			break;

		case 'k':
			file_prio = FILE_PRIO_KEEP;
			break;

		case 'l':
			max_prio = kw_to_num(optarg, strpri, NSTR(strpri));
			if (max_prio == -1) {
				long n;
				char *p;

				errno = 0;
				n = strtol(optarg, &p, 10);
				if (errno || n < 0 || n > LOG_DEBUG)
					error(EX_USAGE, errno, "not a valid number: %s", optarg);
				else if (*p)
					error(EX_USAGE, 0, "not a valid number or priority name");
				max_prio = n;
			}
			break;

		case 'h':
			usage();
			exit(0);
		case 'n':
			/* ignored */
			break;

		case 'p':
			logdev = optarg;
			break;

		default:
			exit(EX_USAGE);
		}
	}

	if (argc > optind)
		error(EX_USAGE, 0, "too many arguments; try %s -h for help", progname);

	if (!hostname)
		hostname = xgethostname();
	hostname_length = strlen(hostname);
	hostname = realloc(hostname, hostname_length + 2);
	if (!hostname)
		enomem();
	hostname[hostname_length] = ' ';
	hostname[hostname_length+1] = 0;

	act.sa_flags = 0;
	act.sa_handler = sighan;
	sigemptyset (&act.sa_mask);
	sigaddset(&act.sa_mask, SIGHUP);
	sigaction(SIGHUP, &act, NULL);
	sigaddset(&act.sa_mask, SIGQUIT);
	sigaction(SIGQUIT, &act, NULL);
	sigaddset(&act.sa_mask, SIGTERM);
	sigaction(SIGTERM, &act, NULL);
	
	fd = open_unix_socket(logdev);
	channel = channel_open(chtype, chname);

	for (;;) {
		ssize_t len;
		char buf[MAXLINE];
		struct message msg;

		len = recv(fd, buf, sizeof(buf), 0);
		if (len < 0) {
			if (errno == EINTR) {
				channel_close(channel);
				if (got_signal == SIGNAL_RELOAD) {
					channel = channel_open(chtype, chname);
				} else
					break;
			} else {
				error(0, errno, "%s", "recv");
			}
			continue;
		}
		if (len == 0) {
			error(0, 0, "ignoring zero-length packet");
			continue;
		}

		parse_message(&msg, buf, len);
		if (PRI_SEV(msg.pri) > max_prio)
			continue;
		message_normalize(&msg);

		channel_send(channel, &msg);
	}
}
